/*
 * Copyright (c) 2005 by G4SRA
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

#include <signal.h>
#include <sys/time.h>

#include "bequest.h"
#include "tcp.h"
#include "omsyslog.h"
#include "omstatus.h"
#include "sys.h"
#include "openmosix.h"
#include "node.h"
#include "semaphore.h"
#include "telinit.h"

typedef struct beqfun {
  char *msg;
  int len;
  int (*fun) (void);
  char *fmt;
  void *arg1;
  void *arg2;
} beqfun_t;

int beqNAInsNode (void);
int beqIPInsNode (void);

int beqNAJoin (void);
int beqIPJoin (void);

int beqNADelNode (void);
int beqIPDelNode (void);

int beqNAPowerDown (void);
int beqIPPowerDown (void);

int beqNAArch (void);
int beqIPArch (void);
int beqNodeNumArch (void);

int beqNAOMVer (void);
int beqIPOMVer (void);
int beqNodeNumOMVer (void);

int beqNAFreeMem (void);
int beqNALoad (void);
int beqNALife (void);
int beqNACPUGrunt (void);
int beqNANetCost (void);
int beqNAMACAddr (void);
int beqNAFlags (void);
int beqNAExit (void);

/* depreciated */

int beqNANumCPU (void);
int beqNASpeed (void);
int beqFubar (void);

static struct in_addr addr;
static int nn;
static omd_node *n;
static char ip[32] = { 0 };
static int dparam;
static double ddparam;
static char sparam[80];

/* order by most frequently used (volatile) first */
/* sizeof()-1 dont count the terminating '\0' */
beqfun_t beqfuntbl[] = {
  {MSG_BEQ_NA_LOAD, (sizeof (MSG_BEQ_NA_LOAD) - 1), beqNALoad, "%08x %lf",
   &addr.s_addr, &ddparam}
  ,
  {MSG_BEQ_NA_FREE_MEM, (sizeof (MSG_BEQ_NA_FREE_MEM) - 1), beqNAFreeMem,
   "%08x %d", &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_INS, (sizeof (MSG_BEQ_NA_INS) - 1), beqNAInsNode, "%08x",
   &addr.s_addr, NULL}
  ,
  {MSG_BEQ_NA_JOIN, (sizeof (MSG_BEQ_NA_JOIN) - 1), beqNAJoin, "%08x",
   &addr.s_addr, NULL}
  ,
  {MSG_BEQ_NA_DEL, (sizeof (MSG_BEQ_NA_DEL) - 1), beqNADelNode, "%08x",
   &addr.s_addr, NULL}
  ,
  {MSG_BEQ_NA_EXIT, (sizeof (MSG_BEQ_NA_EXIT) - 1), beqNAExit, "%08x",
   &addr.s_addr, NULL}
  ,
  {MSG_BEQ_IP_POWERDOWN, (sizeof (MSG_BEQ_IP_POWERDOWN) - 1),
   beqIPPowerDown, "%08x %d",
   &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_POWERDOWN, (sizeof (MSG_BEQ_NA_POWERDOWN) - 1),
   beqNAPowerDown, "%08x %d",
   &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_ARCH, (sizeof (MSG_BEQ_NA_ARCH) - 1), beqNAArch, "%08x %79s",
   &addr.s_addr, sparam}
  ,
  {MSG_BEQ_NA_FLAGS, (sizeof (MSG_BEQ_NA_FLAGS) - 1), beqNAFlags,
   "%08x %d", &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_LIFE, (sizeof (MSG_BEQ_NA_LIFE) - 1), beqNALife,
   "%08x %d", &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_NETCOST, (sizeof (MSG_BEQ_NA_NETCOST) - 1),
   beqNANetCost,
   "%08x %d", &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_MACADDR, (sizeof (MSG_BEQ_NA_MACADDR) - 1), beqNAMACAddr,
   "%08x %79s",
   &addr.s_addr, sparam}
  ,
  {MSG_BEQ_NA_CPUGRUNT, (sizeof (MSG_BEQ_NA_CPUGRUNT) - 1),
   beqNACPUGrunt,
   "%08x %d", &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_SPEED, (sizeof (MSG_BEQ_NA_SPEED) - 1), beqNASpeed,
   "%08x %d", &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_NCPU, (sizeof (MSG_BEQ_NA_NCPU) - 1), beqNANumCPU, "%08x %d",
   &addr.s_addr, &dparam}
  ,
  {MSG_BEQ_NA_OMV, (sizeof (MSG_BEQ_NA_OMV) - 1), beqNAOMVer, "%08x %79s",
   &addr.s_addr, sparam}
  ,
  {MSG_BEQ_IP_ARCH, (sizeof (MSG_BEQ_IP_ARCH) - 1), beqIPArch, "%31s %79s",
   ip,
   sparam}
  ,
  {MSG_BEQ_IP_JOIN, (sizeof (MSG_BEQ_IP_JOIN) - 1), beqIPJoin, "%31s", ip,
   NULL}
  ,
  {MSG_BEQ_IP_INS, (sizeof (MSG_BEQ_IP_INS) - 1), beqIPInsNode, "%31s", ip,
   NULL}
  ,
  {MSG_BEQ_IP_DEL, (sizeof (MSG_BEQ_IP_DEL) - 1), beqIPDelNode, "%31s", ip,
   NULL}
  ,
  {MSG_BEQ_IP_OMV, (sizeof (MSG_BEQ_IP_OMV) - 1), beqIPOMVer, "%31s %79s", ip,
   sparam}
  ,
  {MSG_BEQ_NN_ARCH, (sizeof (MSG_BEQ_NN_ARCH) - 1), beqNodeNumArch, "%d %79s",
   &nn, sparam}
  ,
  {MSG_BEQ_NN_OMV, (sizeof (MSG_BEQ_NN_OMV) - 1), beqNodeNumOMVer, "%d %79s",
   &nn, sparam}
  ,
  {"FUBAR ", (sizeof ("FUBAR ") - 1), beqFubar, NULL, NULL, NULL}
};

omd_node *
omd_get_node_by_netaddr (const struct in_addr *node_addr)
{
  int i;
  omd_node *n;
  uint32_t t_addr = node_addr->s_addr;

  for (i = 0; i < cluster->end; i++) {
    if ((uint32_t) (n = v_get (cluster, i))->s_addr == t_addr) {
/*
#ifdef _DEBUG_
      omLog (OM_LOG_DEBUG, "Identified node %d %s %08x", i,
	       inet_ntoa (*node_addr), ntohl (t_addr));
#endif
*/
      return (n);
    }
  }
  omLog (OM_LOG_NOTICE, "No such node %s (%08x)",
	 inet_ntoa (*node_addr), ntohl (t_addr));
  return (NULL);
}

int
beqNodeNumArch ()
{
  if ((n = v_get (cluster, nn)) != NULL) {
    return (set_node_arch (n, sparam));
  }
  return (OM_STATUS_BAD_NODENUM);
}

int
beqNAArch ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_arch (n, sparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqIPArch ()
{
  inet_pton (AF_INET, ip, &addr);
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_arch (n, sparam));
  }
  return (OM_STATUS_BAD_NODEIP);
}

int
beqNAPowerDown ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    if (n->s_addr == 0) {
      telinit (&dparam);
      return (OM_STATUS_SUCCESS);
    }
    else {
      /* should never happen */
      return (OM_STATUS_NOACTION);
    }
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqIPPowerDown ()
{
  inet_pton (AF_INET, ip, &addr);
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    if (n->s_addr == 0) {
      telinit (&dparam);
      return (OM_STATUS_SUCCESS);
    }
    else {
      /* should never happen */
      return (OM_STATUS_NOACTION);
    }
  }
  return (OM_STATUS_BAD_NODEIP);
}

int
beqNAInsNode ()
{
  /* FIXME: change the arguments */
  if ((n = omd_get_node_by_netaddr (&addr)) == NULL) {
    /* FIXME: change the arguments */
    return (omd_ins_node (&addr, 0, 0));
  }
  else {
    return (OM_STATUS_NOACTION);
  }
}

int
beqIPInsNode ()
{
  inet_pton (AF_INET, ip, &addr);
  return (beqNAInsNode ());
}

int
beqFubar ()
{
  omLog (OM_LOG_ALERT, "Received Fubar bequest");
  return (OM_STATUS_FAIL);
}

int
beqNADelNode ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    /* change the arguments */
    return (omd_del_node (&addr, 0, 0));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqIPDelNode ()
{
  inet_pton (AF_INET, ip, &addr);
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    /* FIXME: change the arguments */
    return (omd_del_node (&addr, 0, 0));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNAJoin ()
{

  if ((n = omd_get_node_by_netaddr (&addr)) == NULL) {
    /* FIXME: change the arguments */
    return (omd_ins_node (&addr, 0, 0));
  }
  else if (n->life == (-1)) {
    /* A powerd down drone has come back up */
    n->life = 0;
    return (OM_STATUS_SUCCESS);
  }
  else {
    return (OM_STATUS_NOACTION);
  }
}

int
beqIPJoin ()
{
  inet_pton (AF_INET, ip, &addr);
  return (beqNAJoin ());
}

int
beqNAExit ()
{
  /* change the arguments */
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    if (n->s_addr != 0) {
      if (n->flags & SYS_FLAGS_POWER) {
	/* FIXME: this gets it wrong after drone restart
	   and drone's flags have changed */
	n->life = (-1);
	return (OM_STATUS_SUCCESS);
      }
      else {
	return (omd_del_node (&addr, 0, 0));
      }
    }
    else {
      /* should never happen */
      return (OM_STATUS_NOACTION);
    }

  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqIPExit ()
{
  inet_pton (AF_INET, ip, &addr);
  return (beqNAExit ());
}

int
beqNAFlags ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_flags (n, &dparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNANetCost ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_netcost (n, &dparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

/* FIXME: finish me */
int
beqNAMACAddr (void)
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_macaddr (n, sparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNACPUGrunt ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_grunt (n, &dparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNASpeed ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_speed (n, &dparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNANumCPU ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_numcpu (n, &dparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNAOMVer ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_omver (n, sparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNALife ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_lifetime (n, &dparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNAFreeMem ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_freemem (n, &dparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqNALoad ()
{
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_load (n, &ddparam));
  }
  return (OM_STATUS_BAD_NODEADDR);
}

int
beqIPOMVer ()
{
  inet_pton (AF_INET, ip, &addr);
  if ((n = omd_get_node_by_netaddr (&addr)) != NULL) {
    return (set_node_omver (n, sparam));
  }
  return (OM_STATUS_BAD_NODEIP);
}

int
beqNodeNumOMVer ()
{
  if ((n = v_get (cluster, nn)) != NULL) {
    return (set_node_omver (n, sparam));
  }
  return (OM_STATUS_BAD_NODENUM);
}

/* handle local stuff */
int
handle_bequest (char *buf)
{
  int i, ret;

  for (i = 0; i < sizeof (beqfuntbl) / sizeof (beqfun_t); i++) {
    if (!strncmp (buf, beqfuntbl[i].msg, beqfuntbl[i].len)) {
#ifdef _DEBUG_
      omLog (OM_LOG_DEBUG, "Received bequest \"%s\"", buf);
#endif
      sscanf (&buf[beqfuntbl[i].len], beqfuntbl[i].fmt, beqfuntbl[i].arg1,
	      beqfuntbl[i].arg2);
      down_write (&cluster_semaphore);
      ret = beqfuntbl[i].fun ();
      up_write (&cluster_semaphore);
      return (ret);
    }
  }
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "handle_bequest \"%s\" unknown", buf);
#endif
  return (OM_STATUS_BAD_BEQ);
}

#if 0
int
handle_bequest_1 (char *buf)
{
  char *ptr1, *ptr2;
  int ret = OM_STATUS_FAIL;

  ptr2 = buf;
  while (ptr2 != NULL) {
    ptr1 = ptr2;
    ptr2 = strchr (ptr1, '\n');
    if (ptr2 != NULL) {
      /* we have a multiple command string */
      *ptr2 = '\0';
      ++ptr2;
      if (*ptr2 == '\0') {
	ptr2 = NULL;
      }
    }
    ret = handle_bequest (ptr1);
#ifdef _DEBUG_
/*
    if (ret < 0) {
      omLog (OM_LOG_DEBUG, "handle_bequest [%s] for \"%s\"",
	       strstatus (ret), buf);
    }
*/
#endif
  }
  return (ret);
}
#endif
