/*
* Copyright (c) 2005 by G4SRA
*
* Permission to use, copy and distribute this software is hereby granted
* under the terms of version 2 or any later version of the GNU General Public
* License, as published by the Free Software Foundation.
*
* THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
* WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
* FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
*/
/*
* Authors: G4SRA (cradford@bcs.org.uk)
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <dirent.h>
#include <sys/time.h>
#include <values.h>

#include "omd.h"
#include "omsyslog.h"
#include "omstatus.h"
#include "openmosix.h"
#include "semaphore.h"
#include "request.h"
#include "bequest.h"
#include "tcp.h"
#include "sys.h"


int
omCompatible (int nn)
{
  int ret = 0;
  omd_node *n;
  /* check that remote node (architecture, omv, etc) is compatible */
#ifdef _ROBUST_
  if (nn < 0) {
    omLog (OM_LOG_NOTICE, "ROBUST omCompatible %d", nn);
    return (0);
  }
  if (nn != 0) {
#endif
    down_read (&cluster_semaphore);
    if ((n = v_get (cluster, nn)) != NULL) {
      ret = omNodeCompatible (v_get (cluster, 0), n);
    }
    up_read (&cluster_semaphore);
#ifdef _ROBUST_
  }
  else {
    omLog (OM_LOG_NOTICE, "ROBUST omCompatible node 0 is always compatible");
    ret = MAXINT;
  }
#endif
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "omCompatible node:%d \"%s\"", nn,
	 (ret ? "Yes" : "No"));
#endif
  return (ret);
}

int
omMigrate (int pid, int nn)
{
  struct in_addr addr;
  char str1[256];
  char str2[256];
  int fd;
  int tmp;
  omd_node *n;

  omLog (OM_LOG_NOTICE, "Migrating pid %d to node %d", pid, nn);

  snprintf (str1, sizeof (str1) - 1, "/proc/%d/om/where", pid);

#ifdef _ROBUST_
  if ((unsigned int) nn > omd_cluster_size ()) {
    omLog (OM_LOG_NOTICE, "ROBUST omMigrate node %d unknown", nn);
    return (OM_STATUS_FAIL);
  }
#endif

  n = v_get (cluster, nn);

#ifdef _ROBUST_
  if (n == NULL) {
    omLog (OM_LOG_NOTICE, "ROBUST omMigrate node is NULL");
    return (OM_STATUS_FAIL);
  }

  if (n->s_addr < 0) {
    omLog (OM_LOG_NOTICE, "ROBUST omMigrate node bad address");
    return (OM_STATUS_FAIL);
  }

  if ((fd = open (str1, O_RDONLY)) < 0) {
    omLog (OM_LOG_CRIT, "ROBUST omMigrate open for reading \"%s\":%s", str1,
	   strerror (errno));
    return (OM_STATUS_FAIL);
  }

  if ((tmp = read (fd, str2, sizeof (str2))) < 0) {
    omLog (OM_LOG_CRIT, "ROBUST omMigrate reading:%s", strerror (errno));
    close (fd);
    return (OM_STATUS_FAIL);
  }

  close (fd);
  str2[tmp] = '\0';

  if (strncmp ("home\n", str2, strlen (str2)) != 0) {
    omLog (OM_LOG_WARNING,
	   "ROBUST omMigrate not at home \"%s\" strlen=%d %d", str2,
	   strlen (str2), tmp);
    return (OM_STATUS_FAIL);
  }
#endif

  addr.s_addr = n->s_addr;
  snprintf (str2, sizeof (str2) - 1, "%s", inet_ntoa (addr));

  if ((fd = open (str1, O_RDWR | O_TRUNC)) < 0) {
    omLog (OM_LOG_CRIT, "omMigrate open for writing:%s", strerror (errno));
    return (OM_STATUS_FAIL);
  }

  if (write (fd, str2, strlen (str2)) < 0) {
    close (fd);
    omLog (OM_LOG_CRIT, "omMigrate writing:%s", strerror (errno));
    return (OM_STATUS_FAIL);
  }
  close (fd);
  return (OM_STATUS_SUCCESS);
}

int
omCandidateDrone (double load_avg)
{
  int i;
  int chosen_n = (-1);
  omd_node *n, *p;

  p = v_get (cluster, 0);
  for (i = cluster->end - 1; i > 0; i--) {
#ifdef _DEBUG_
    omLog (OM_LOG_DEBUG, "Examining node %d of %d", i, cluster->end);
#endif
    n = v_get (cluster, i);
    if ((n->life >= 0) && (n->flags & SYS_FLAGS_DRONE)) {
      if (n->load_avg < 0) {
	struct in_addr addr;
	/*
	   double load_avg = (-1);
	   set_node_load(n,&load_avg);
	 */
#ifdef _DEBUG_
	omLog (OM_LOG_DEBUG, "Drone node %d has stale load %lf", i,
	       n->load_avg);
#endif

#ifdef _FORKED_REQUESTER_
	if (fork () == 0) {
#endif
#if 0
	  {
	    int sd1, sd2;
	    char *buf;
	    if ((sd1 = tcp_chat_start (n->s_addr)) != -1) {
	      tcp_send (&sd1, MSG_REQ_IP_LOAD "0.0.0.0");
	      if ((buf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
		if ((sd2 = tcp_chat_start (INADDR_ANY)) != -1) {
		  tcp_sendf (&sd2, MSG_BEQ_NA_LOAD "%08x %s", n->s_addr, buf);
		  close (sd2);
		}
#else
		sscanf (buf, "%lf", &load_avg);
#endif
		free (buf);
	      }
	      close (sd1);
	    }
	  }
#else
	  addr.s_addr = n->s_addr;
	  omd_volatile ((&addr));
#endif
#ifdef _FORKED_REQUESTER_
	  exit (0);
	}
#endif

	/* FIXME: this is always going to be true */
	if (n->load_avg < 0) {
#ifdef _DEBUG_
	  omLog (OM_LOG_DEBUG, "Skipping node %d", i);
#endif
	  continue;
	}
      }
#ifdef _DEBUG_
      omLog (OM_LOG_DEBUG, "Drone node %d load:%lf", i, n->load_avg);
#endif
      /* change this argument to om_node */
      if (omCompatible (i)) {
	/* need to consider load, free memory, mips, network cost */
	/* this order is all very subjective */
	if (n->load_avg <= p->load_avg) {
	  if (p->load_avg < 0.25) {
	    /* currently selected node load average is lower than 0.25
	     * cpu power & network cost will now have greater effect  */
	    if (n->grunt > p->grunt) {
	      /* load <=, grunt >, */
	      p = n;
	      chosen_n = i;
	    }
	    else if (n->grunt == p->grunt) {
	      if (n->net_cost <= p->net_cost) {
		/* load <=, grunt =, net_cost <=, */
		p = n;
		chosen_n = i;
	      }
	      /* load <=, grunt =, net_cost >, */
	    }
	    /* load <=, grunt <, */
	  }
	  else {
	    /* load <=, */
	    p = n;
	    chosen_n = i;
	  }
	}
      }
    }
  }
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "Chosen drone node:%d", chosen_n);
#endif
  return (chosen_n);
}

float
omElapsedTime (void)
{
  static struct timeval oldtime;
  struct timeval time;
  struct timezone timez;
  float elapsed_time;

  gettimeofday (&time, &timez);
  elapsed_time = ((float) (time.tv_sec - oldtime.tv_sec))
    + (((float) (time.tv_usec - oldtime.tv_usec)) / 1000000.0);

/*
  elapsed_time = (((float)(time.tv_sec - oldtime.tv_sec)) * 1000000.0)
      + ((float) (time.tv_usec - oldtime.tv_usec));
*/

  oldtime.tv_sec = time.tv_sec;
  oldtime.tv_usec = time.tv_usec;
  return (elapsed_time);
}

      /* this is all om kernel dependant, have to
         wait until the developers release/finish it */
pid_t
omCandidateProcess (void)
{
  pid_t chosen_pid = (-1);
  unsigned long chosen_time = 0;
  unsigned long total_time;
  unsigned long elapsed_time;
  DIR *pt;
  struct dirent *dir;
  char path[64];
  int fd;
  int num_read;
  char buf[256];
  char *tmp;
  proc_t *pr, pro;
  struct stat sb;

  if ((pt = opendir ("/proc")) != NULL) {
    pr = &pro;
    chosen_pid = 0;
    elapsed_time = omElapsedTime ();
    while ((dir = readdir (pt)) != NULL) {
      if (dir->d_name[0] < '0' || dir->d_name[0] > '9')
	continue;

#ifndef _MIGRATE_ROOT_PROCS_
      /* do not migrate anything running as root */
      /* this could be changed to uid,gid < 500 */
      sprintf (path, "/proc/%s", dir->d_name);
      stat (path, &sb);
      if ((sb.st_uid == 0) || (sb.st_gid == 0)) {
#ifdef _DEBUG_
	omLog (OM_LOG_DEBUG, "Rejecting pid:%s, uid:%d gid:%d",
	       dir->d_name, sb.st_uid, sb.st_gid);
#endif
	continue;
      }
#endif

      /* stat() always returns 0 file size for /proc/???/om/stay
         so we have to read and count the bytes */
      sprintf (path, "/proc/%s/om/stay", dir->d_name);
      if ((fd = open (path, O_RDONLY, 0)) == -1) {
	omLog (OM_LOG_CRIT, "Cannot open %s:%s", path, strerror (errno));
	continue;
      }
      num_read = read (fd, buf, sizeof (buf) - 1);
      close (fd);
      if (num_read < 0) {
	omLog (OM_LOG_CRIT, "Cannot read %s:%s", path, strerror (errno));
	continue;
      }
      buf[num_read] = '\0';
      if (num_read > 0) {
	/* any reason to stay is a valid reason not to migrate */
#ifdef _DEBUG_
	omLog (OM_LOG_DEBUG, "Rejecting pid:%s, reason to stay \"%s\"",
	       dir->d_name, buf);
#endif
	continue;
      }

      sprintf (path, "/proc/%s/om/where", dir->d_name);
      if ((fd = open (path, O_RDONLY, 0)) == -1) {
	omLog (OM_LOG_CRIT, "Cannot open %s:%s", path, strerror (errno));
	continue;
      }
      num_read = read (fd, buf, sizeof (buf) - 1);
      close (fd);
      if (num_read <= 0) {
	omLog (OM_LOG_CRIT, "Cannot read %s:%s", path, strerror (errno));
	continue;
      }
      buf[num_read] = '\0';
      if (strncmp ("home\n", buf, num_read)) {
#ifdef _DEBUG_
	omLog (OM_LOG_DEBUG, "Rejecting pid:%s, not home \"%s\"",
	       dir->d_name, buf);
#endif
	continue;
      }

      sprintf (path, "/proc/%s/stat", dir->d_name);
      if ((fd = open (path, O_RDONLY, 0)) == -1) {
	omLog (OM_LOG_CRIT, "Cannot open %s:%s", path, strerror (errno));
	continue;
      }
      num_read = read (fd, buf, sizeof (buf) - 1);
      close (fd);
      if (num_read <= 0) {
	omLog (OM_LOG_CRIT, "Cannot read %s:%s", path, strerror (errno));
	continue;
      }
      buf[num_read] = '\0';


      /* Might as well sscanf the pid, could use dir->d_name */

      /* sscanf (buf, "%d (%s) ", &pr->pid, pr->tcomm); */
      /* sscanf (or I) has trouble with bracketed spaced executable names */
      sscanf (buf, "%d ", &pr->pid);
/*
      tmp1 = strchr (buf, '(');
      tmp2 = strchr (buf, ')');
      tmp1++;
      while(tmp1 < tmp2) {
        *buf++ = *tmp1++
      }
      sscanf (tmp+1, "%s)", pr->tcomm);
*/
      /* more than one way to sscanf a cat */
      tmp = strchr (buf, ')');
      sscanf (tmp + 2, "%c ", &pr->state);
      if (pr->state != 'R') {
#ifdef _DEBUG_
	omLog (OM_LOG_DEBUG, "Rejecting pid:%s, status '%c' not 'R'",
	       dir->d_name, pr->state);
#endif
	continue;
      }

      sscanf (tmp + 4, "%d %d %d %d %d \
%lu %lu %lu %lu %lu \
%lu %lu %ld %ld \
%ld %ld %d %ld %lu %lu %lu \
%lu %lu %lu %lu %lu \
%lu %lu %lu %lu %lu %lu %lu %lu %d %d %lu %lu\n", &pr->ppid, &pr->pgid, &pr->sid, &pr->tty_nr, &pr->tty_pgrp, &pr->task_flags, &pr->min_flt, &pr->cmin_flt, &pr->maj_flt, &pr->cmaj_flt, &pr->utime, &pr->stime, &pr->cutime, &pr->cstime, &pr->priority, &pr->nice, &pr->num_threads, &pr->it_real_value, &pr->start_time, &pr->vsize, &pr->mm_counter, &pr->rsslim, &pr->mm_start_code, &pr->mm_end_code, &pr->mm_start_stack, &pr->esp, &pr->eip, &pr->task_pendingsig, &pr->task_blockedsig, &pr->sigignsig, &pr->sigcatchsig, &pr->wchan, &pr->junk1, &pr->junk2, &pr->task_exit_signal, &pr->task_cpu, &pr->task_rt_priority, &pr->task_policy);

#ifdef _ROBUST_
      /* for now, only migrate anything niced to 1 */
      if (pr->nice != 1) {
	/* high prioirity, we need to see this when it happens */
	omLog (OM_LOG_ALERT, "ROBUST Rejecting pid:%s, niceness:%d",
	       dir->d_name, pr->nice);
	continue;
      }
#endif

      /* FIXME: very dodgy code :>) */
      /* (pr->utime + pr->stime) are not in sync with elapsed_time timewise */
      total_time = (pr->utime + pr->stime) / elapsed_time;
      /* pr->stime is usually so small as to be insignificant */
      if (total_time >= chosen_time) {
#ifdef _DEBUG_
	omLog (OM_LOG_DEBUG, "Accept pid:%s, (%lu+%lu)/%ld=%ld >= %ld",
	       dir->d_name, pr->utime, pr->stime, elapsed_time, total_time,
	       chosen_time);
#endif
	chosen_time = total_time;
	chosen_pid = pr->pid;
      }
#ifdef _DEBUG_
      else {
	omLog (OM_LOG_DEBUG, "Rejecting pid:%s, (%lu+%lu)/%ld=%ld >= %ld",
	       dir->d_name, pr->utime, pr->stime, elapsed_time, total_time,
	       chosen_time);
      }
#endif
    }
  }
  else {
    omLog (OM_LOG_CRIT, "Cannot open /proc:%s", strerror (errno));
  }
  return (chosen_pid);
}
