/*
* Copyright (c) 2005 by G4SRA
*
* Permission to use, copy and distribute this software is hereby granted
* under the terms of version 2 or any later version of the GNU General Public
* License, as published by the Free Software Foundation.
*
* THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
* WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
* FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
*/
/*
* Authors: G4SRA (cradford@bcs.org.uk)
*/


#include <unistd.h>
#include <stdlib.h>
#include <sys/utsname.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <values.h>

#include "node.h"
#include "omstatus.h"
#include "omlocal.h"
#include "libmisc.h"
#include "ommacaddr.h"

omd_node *
omNodeCreate (uint32_t addr)
{

  /* omd_node *n = (omd_node *) omMalloc (sizeof (omd_node)); */
  omd_node *n = (omd_node *) malloc (sizeof (omd_node));
  if (n != NULL) {

    /* memset (n, '\0', sizeof (omd_node)); */
    n->s_addr = addr;
    n->alias = (-1);
    n->alias_cnt = (-1);
    n->num_cpus = (-1);
    n->grunt = (-1);
    n->net_cost = (-1);
    n->arch = NULL;
    n->mips = NULL;
/*    n->ver = NULL; */
    memset (n->macaddr, 0, 6);
    n->free_mem = (-1);
    n->load_avg = (-1);
    n->lifetime = (-1);
    n->life = (-1);
    n->flags = (-1);
    n->mver = 0;
    n->rel = '\0';
    n->rev = 0;
  }
  return n;
}

omd_node *
omNodeDestroy (omd_node ** n)
{
  if ((n != NULL) && (*n != NULL)) {
    if ((*n)->arch != NULL) {
      free ((*n)->arch);
      (*n)->arch = NULL;
    }
/*
    if ((*n)->ver != NULL) {
      free ((*n)->ver);
      (*n)->ver = NULL;
    }
*/
    if ((*n)->mips != NULL) {
      free ((*n)->mips);
      (*n)->mips = NULL;
    }
    free (*n);
    *n = NULL;
    return (*n);
  }
  return (NULL);
}

int
set_node_arch (omd_node * n, char *arch)
{
  if (n->arch != NULL)
    free (n->arch);
  n->arch = omStrdup (arch);
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_flags (omd_node * n, int *flags)
{
  n->flags = *flags;
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_omver (omd_node * n, char *omver)
{
  unsigned char mver, rel, rev;

  mver = rev = 0;
  while ((*omver >= '0') && (*omver <= '9')) {
    mver *= 10;
    mver += *omver - '0';
    ++omver;
  }
  rel = *omver++;
  while ((*omver >= '0') && (*omver <= '9')) {
    rev *= 10;
    rev += *omver - '0';
    ++omver;
  }

  n->mver = mver;
  n->rel = rel;
  n->rev = rev;

  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

char *
get_node_omver (omd_node * n)
{
  static char omver[64];

  if (n->rel != 0) {
    sprintf (omver, "%d%c%d", n->mver, n->rel, n->rev);
    return (omver);
  }
  return (NULL);
}

int
set_node_macaddr (omd_node * n, char *macaddr)
{
  unsigned char *ptr;

  ptr = omStrMACAddr ((unsigned char *) macaddr);
  if (ptr != NULL)
    memcpy (n->macaddr, ptr, 6);
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_speed (omd_node * n, int *mips)
{
  if (n->mips != NULL)
    free (n->mips);
  n->mips = omMalloc (sizeof (int));
  n->mips[0] = *mips;
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_netcost (omd_node * n, int *net_cost)
{
  n->net_cost = *net_cost;
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_grunt (omd_node * n, int *grunt)
{
  n->grunt = *grunt;
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_numcpu (omd_node * n, int *ncpu)
{
  n->num_cpus = *ncpu;
  return (OM_STATUS_SUCCESS);
}

int
set_node_freemem (omd_node * n, int *frmem)
{
  n->free_mem = *frmem;
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_load (omd_node * n, double *loadavg)
{
  n->load_avg = *loadavg;
  n->life = n->lifetime;
  return (OM_STATUS_SUCCESS);
}

int
set_node_lifetime (omd_node * n, int *life)
{
  n->lifetime = *life;
  return (OM_STATUS_SUCCESS);
}

/* return the major version */
/* NOTE: version number is 0 based */
int
omNodeHeirVer (omd_node * n)
{
  int ret = (int) n->mver;
/*
  int ret = (-1);
  int val;

  if (n->ver != NULL) {
    if (sscanf (n->ver, "%d.%d", &ret, &val) != 2)
      return (-1);
  }
*/
  return (ret);
}

int
omNodeHeirArch (omd_node * n)
{
  int ret = (-1);

  if (n->arch != NULL) {
    if (!strcmp ("i386", n->arch))
      ret = 1;
    else if (!strcmp ("i486", n->arch))
      ret = 2;
    else if (!strcmp ("i586", n->arch))
      ret = 3;
    else if (!strcmp ("i686", n->arch))
      ret = 4;
  }
  return (ret);
}

/* is node n2 (better than or the same as) compatible with n1 */
/* FIXME: need to check more */
int
omNodeCompatible (omd_node * n1, omd_node * n2)
{
  int ret;

  /* if n1 is better than n2, or either n1 or n2 is unknown, return fail */
  if ((ret = omNodeHeirVer (n1)) > omNodeHeirVer (n2) || ret < 0)
    return (0);
  if ((ret = omNodeHeirArch (n1)) > omNodeHeirArch (n2) || ret < 0)
    return (0);
  return (MAXINT);
}

/***************************************************************************/
/* Local node only functions */

void
omNodeUpdate (omd_node * n)
{
  n->load_avg = getLocalLoadAvg ();
  n->free_mem = getLocalFreeMem ();
  n->life = n->lifetime;
}

void
omNodeInit (omd_node * n)
{
  if (n != NULL) {
    n->mver = 0;
    n->rel = 'a';
    n->rev = 1;
/*    n->ver = getLocalOMLibVer (); */
    n->arch = getLocalArch ();
    n->grunt = getLocalCPUGrunt ();
    n->num_cpus = getLocalCPUMips (&n->mips);
    getLocalMACAddr (n->macaddr);
    n->lifetime = 0;
    n->life = 0;
    n->flags = 0;
    n->s_addr = 0;
    n->net_cost = 0;
  }
}

/***************************************************************************/
