/*
* Copyright (c) 2005 by G4SRA
*
* Permission to use, copy and distribute this software is hereby granted
* under the terms of version 2 or any later version of the GNU General Public
* License, as published by the Free Software Foundation.
*
* THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
* WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
* FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
*/
/*
* Authors: G4SRA (cradford@bcs.org.uk)
*/


#ifndef _OM_NODE_H
#define _OM_NODE_H

#include <stdint.h>

typedef struct {
  uint32_t s_addr;
  uint32_t alias;
  int alias_cnt;
  /* Host node OM Version */
  unsigned char mver;
  unsigned char rel;
  unsigned char rev;

  /* char *ver; */
  char *arch;			/* Host node Architecture type */
  int *mips;
  int num_cpus;
  int grunt;			/* Effective CPU power (higher = better) */

  int net_cost;			/* Effective Network Cost  (lower = better) */
  int flags;
  double load_avg;		/* Host node average load */
  int free_mem;			/* Host node free memory */
  int lifetime;			/* Cache time to live */

  int life;			/* Cache life */
  unsigned char macaddr[6];
} omd_node;


omd_node *omNodeCreate (uint32_t addr);
omd_node *omNodeDestroy (omd_node ** n);
int omNodeCompatible (omd_node * n1, omd_node * n2);

void omNodeInit (omd_node * n);
void omNodeUpdate (omd_node * n);

int set_node_omver (omd_node * n, char *);
char *get_node_omver (omd_node * n);

int set_node_arch (omd_node * n, char *);
int set_node_grunt (omd_node * n, int *);

int set_node_lifetime (omd_node * n, int *);
int set_node_netcost (omd_node * n, int *);
int set_node_flags (omd_node * n, int *);
int set_node_freemem (omd_node * n, int *);
int set_node_load (omd_node * n, double *);
int set_node_macaddr (omd_node * n, char *macaddr);

/* depreciated */

int set_node_numcpu (omd_node * n, int *);
int set_node_speed (omd_node * n, int *);



#endif
