/*
    name : omlocal.c

    omlib : openMosix Userland Tools Library

    Copyright (C) 2005 Pierre Pacaud, Thomas Roux, Matthieu Pérotin
        Ecole Polytechnique de l'Université de Tours, Département Informatique
    Copyright (C) 2005 Evan Hisey, Nathan Wiegand, G4SRA
	
    omlib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING . If not, write to the
    Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <stdlib.h>
#include <sys/utsname.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <linux/sockios.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <fcntl.h>

#include "omd.h"
#include "omlocal.h"
#include "node.h"
#include "libmisc.h"

/* FIXME: - I hate this  */
ssize_t getline (char **lineptr, size_t * n, FILE * stream);


/* Gets load average in one minute windows for localhost */
double
getLocalLoadAvg (void)
{
  double load_avg[1] = { -1 };

  if (getloadavg (load_avg, 1) == -1) {
    /* for some reason, load averages couldn't be obtained.  */
    return -1;
  }
  /* This is the load average over the last minute,
   * change the parameter to getloadavg to get 
   * longer average windows.
   */
  return load_avg[0];
}

/* Gets current free memory of local host. */
int
getLocalFreeMem (void)
{
  FILE *fp;
  char *buffer = NULL;
  size_t len = 0;
  ssize_t read;

  fp = fopen ("/proc/meminfo", "r");

  if (!fp) {
    /* Could not open /proc/meminfo, 
     * something horrible must have happened.
     */
    return -1;
  }

  while ((read = getline (&buffer, &len, fp)) != -1) {
    /* Looking for the MemFree line. */
    char unit[4] = { 0 };
    int value;

    if (!strncmp ("MemFree:", buffer, 8)) {
      sscanf (buffer, "MemFree: %d %s", &value, unit);
      /* Might want to eventually compare 'unit'
       * to kB, mB, etc., just in case that changes.
       */
      fclose (fp);
      return value;
    }
  }
  /* For some reason, finding the value from 
   * /proc/meminfo didn't work.
   * You really should call someone about this.
   */
  fclose (fp);
  return -1;
}

  /* Not all Linux systems have their CPU speed recorded in /proc/cpuinfo
   * and what does Mhz tell us anyway, Mips do have meaning, and are there.
   */
int
getLocalCPUMips (int **mips)
{

  FILE *fp;
  char *buffer = NULL;
  size_t len = 0;
  ssize_t read;

  int number_cpus = 0, i;

  /* does ANYONE have more than 124 local cpus?
   * it seems like the speed gain of having this on the
   * stack will be worth it.
   */
  int freqs[124] = { -1 }, *retn_array;

  /* Something cool could be done here with sysctl, 
   * but it seems to be sparsely available in the propper form
   */

  /* Linux can simply use bogomips.
   * Other OS'es can always calculate/approximate
   */
  fp = fopen ("/proc/cpuinfo", "r");

  if (!fp) {
    /* Could not open /proc/cpuinfo,
     * something horrible must have happened.
     */
    return -1;
  }

  while ((read = getline (&buffer, &len, fp)) != (-1)) {
    if (!strncmp ("bogomips\t:", buffer, 10)) {
      /* should test to make sure that number_cpus is bounded */
      sscanf (buffer, "bogomips\t: %d", &freqs[number_cpus++]);
    }
  }
  fclose (fp);
  /* retn_array = (int *) omd_malloc (sizeof (int) * number_cpus); */
  retn_array = (int *) malloc (sizeof (int) * number_cpus);
  if (retn_array == NULL)
    return (-1);
  for (i = 0; i < number_cpus; i++) {
    retn_array[i] = freqs[i];
  }
  /* return the mips and the number of CPUS */
  *mips = retn_array;
  return number_cpus;
}

int
getLocalNumCPUs (void)
{
  FILE *fp;
  char *buffer = NULL;
  size_t len = 0;
  ssize_t read;

  int number_cpus = 0;

  fp = fopen ("/proc/cpuinfo", "r");
  if (!fp) {
    /* Could not open /proc/cpuinfo,
     * something horrible must have happened.
     */
    return -1;
  }

  while ((read = getline (&buffer, &len, fp)) != (-1)) {
    if (!strncmp ("processor\t:", buffer, 11)) {
      ++number_cpus;
    }
  }
  fclose (fp);
  return number_cpus;
}

char *
getLocalArch (void)
{
  struct utsname buf;

  if (uname (&buf) != (-1)) {
    return (omStrdup (buf.machine));
  }
  return NULL;
}

char *
getLocalKernelVer (void)
{
  int fp;
  char buf[256];

  fp = open ("/proc/hpc/admin/version", O_RDONLY);
  if (fp) {
    read (fp, buf, sizeof (buf) - 1);
    return (omStrdup (buf));
  }
  fp = open ("/proc/version", O_RDONLY);
  if (fp) {
    read (fp, buf, sizeof (buf) - 1);
    return (omStrdup (buf));
  }
  return NULL;
}

char *
getLocalOMLibVer (void)
{
  return (omStrdup (OMD_VERSION));
}

int
getLocalCPUGrunt (void)
{
  int *mips;
  int num_cpus;
  int grunt;

  grunt = 0;
  num_cpus = getLocalCPUMips (&mips);
  while (num_cpus > 0)
    grunt += mips[--num_cpus];
  free (mips);
  grunt >>= 2;
  return (grunt);
}

int
getLocalMACAddr (unsigned char *mac)
{
  struct ifreq ifr;
  int sd = (-1);
  int ret = (-1);

  if ((sd = socket (PF_INET, SOCK_DGRAM, 0)) >= 0) {
    memset (&ifr, 0, sizeof (struct ifreq));
    /* FIXME: hard coded if name is a no-no */
    strcpy (ifr.ifr_name, "eth0");
    if ((ret = ioctl (sd, SIOCGIFHWADDR, &ifr)) >= 0) {
      memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);
      return (ret);
    }
  }
  return (sd);
}

int
openmosix_check (void)
{
  if (access ("/proc/hpc", R_OK | X_OK)) {
    return (-1);
  }
  return (0);
}
