/*
* Copyright (c) 2005 by G4SRA
*
* Permission to use, copy and distribute this software is hereby granted
* under the terms of version 2 or any later version of the GNU General Public
* License, as published by the Free Software Foundation.
*
* THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
* WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
* FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
*
* Author: G4SRA (cradford@bcs.org.uk)
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include "omsyslog.h"
#include "omstatus.h"
#include "ommacaddr.h"

/* Convert MAC Address Array to a Text String */
char *
omMACAddrStr (unsigned char *macptr)
{
  static char macstr[18];

  if (macptr != NULL) {
    sprintf ((char *) macstr, "%02X:%02X:%02X:%02X:%02X:%02X",
	     macptr[0], macptr[1], macptr[2], macptr[3], macptr[4],
	     macptr[5]);
    return (macstr);
  }
  return (NULL);
}

/* Convert MAC Address Text String to an Array */
unsigned char *
omStrMACAddr (unsigned char *strptr)
{
  static unsigned char arry[6];
  int i;

  if (strptr != NULL) {
    memset (arry, 0, 6);
    for (i = 0; i < 6; i++) {
/* FIXME: not robust */
      if (*strptr <= '9')
	arry[i] += *strptr - '0';
      else
	arry[i] += *strptr - ('A' - 10);
      arry[i] <<= 4;
      ++strptr;
      if (*strptr <= '9')
	arry[i] += *strptr - '0';
      else
	arry[i] += *strptr - ('A' - 10);
      ++strptr;
      ++strptr;
    }

/*
    sscanf ((char *) strptr, "%02X:%02X:%02X:%02X:%02X:%02X",
	    &arry[0], &arry[1], &arry[2], &arry[3], &arry[4], &arry[5]);
*/
    return (arry);
  }
  memset (arry, 0, 6);
  return (arry);
  /* return (NULL); */
}

#ifdef _TESTMACADDR_
#include <sys/socket.h>
#include <linux/sockios.h>
#include <sys/ioctl.h>
#include <net/if.h>

unsigned char *
omGetMACAddr (char *host)
{
  static unsigned char macaddr[64];
  static unsigned char *macptr = NULL;
  int linecount = 0;
  int fd;
  char buf[512];
  int num;
  char *ptr;
  int bufcnt;
  char ip[16];
  unsigned int type;
  unsigned int flags;
  char mask[64];
  char dev[64];

  if ((fd = open ("/proc/net/arp", O_RDONLY, 0)) < 0) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "omGetMACAddr:%s", strerror (sys_errno));
    return (NULL);
  }

  ptr = buf;
  bufcnt = 0;
  while (macptr == NULL) {

    while ((num = read (fd, ptr, 1)) < 0) {
      sys_errno = errno;
      if (sys_errno == EINTR) {
	continue;
      }
      omLog (OM_LOG_CRIT, "omGetMACAddr:%s", strerror (sys_errno));
      return (NULL);
    }

    if ((*ptr != '\n') && (num > 0)) {
      if (++bufcnt >= sizeof (buf) - 1) {
	omLog (OM_LOG_ALERT, "omGetMACAddr: buffer overflow");
	return (NULL);
      }
      ++ptr;
      continue;
    }

    if ((bufcnt == 0) && (num == 0)) {
#ifdef _TESTMACADDR_
      omLog (OM_LOG_CRIT, "omGetMACAddr: end of file");
#endif
      break;
    }

    if (!linecount++) {
#ifdef _TESTMACADDR_
      omLog (OM_LOG_CRIT, "omGetMACAddr: skipping header");
#endif
      ptr = buf;
      bufcnt = 0;
      continue;
    }

    ++ptr;
    *ptr = '\0';

    ptr = buf;
    bufcnt = 0;

    num = sscanf (buf, "%s 0x%x 0x%x %64s %64s %64s\n",
		  ip, &type, &flags, macaddr, mask, dev);
#ifdef _TESTMACADDR_
    omLog (OM_LOG_DEBUG, "%s 0x%x 0x%x %s %s %s",
	   ip, &type, &flags, macaddr, mask, dev);
#endif
    if (num < 4) {
      continue;
    }
    if (type != 1) {
      continue;
    }
    if (!strcmp (host, ip)) {
      macptr = macaddr;
      break;
    }
  }
  close (fd);
  return (macptr);
}

int
getLocalMACAddr (unsigned char *mac)
{
  int ctl_sk = -1;
  struct ifreq ifr;
  int r;

  ctl_sk = socket (PF_INET, SOCK_DGRAM, 0);
  memset (&ifr, 0, sizeof (struct ifreq));
  /* FIXME: hard coded if name is a no-no */
  strcpy (ifr.ifr_name, "eth0");
  if ((r = ioctl (ctl_sk, SIOCGIFHWADDR, &ifr)) >= 0) {
    memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);
  }
  return r;
}

int
main (int argc, char *argv[])
{
  unsigned char ary[6];
  unsigned char *ptr;
  int i;

  omLogInitialise (1, "testmac");
  omLogSetMask (255);
  ptr = omGetMACAddr ("192.168.1.69");
  /*ptr = omGetMACAddr ("170.71.233.211"); */
  omLog (OM_LOG_CRIT, "%s", ptr);
/*
  ptr = omStrMACAddr (ptr);
  if (ptr != NULL) {
    memcpy (ary, ptr, 6);
    for (i = 0; i < 6; i++)
      omLog (OM_LOG_CRIT, "%02X", ary[i]);
  }
*/

  i = getLocalMACAddr (ary);
  omLog (OM_LOG_CRIT, "Result=%d", i);
  ptr = omMACAddrStr (ary);
  omLog (OM_LOG_CRIT, "%s", ptr);
  ptr = omStrMACAddr (ptr);
  for (i = 0; i < 6; i++)
    omLog (OM_LOG_CRIT, "%02X", ptr[i]);

  ptr = omMACAddrStr (ptr);
  omLog (OM_LOG_CRIT, "%s", ptr);
  omLogFinalise ();
  return (0);
}
#endif
