#include <sys/param.h>
#include <netinet/ip_icmp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <values.h>

#include "omsyslog.h"
#include "omstatus.h"
#include "sys.h"

#define	MAXIPLEN	60
#define	MAXICMPLEN	76
#define	MAXWAIT		5	/* max seconds to wait for response */

#define	DEFDATALEN	(64 - 8)	/* default data length */
#define	MAXPACKET	(65536 - 60 - 8)	/* max packet size */


int
icmp_socket (int *sd)
{
  struct protoent *proto;
  if (!(proto = getprotobyname ("icmp"))) {
    omLog (OM_LOG_WARNING, "unknown protocol \"icmp\"");
    return (OM_STATUS_FAIL);
  }

  if ((*sd = socket (AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
    if (errno == EPERM) {
      if (getuid () != 0)
	omLog (OM_LOG_WARNING, "must run as root to open socket");
      else
	omLog (OM_LOG_WARNING, "invalid permissions for socket");
    }
    else {
      sys_errno = errno;
      omLog (OM_LOG_CRIT, "socket \nperror: %s", strerror (errno));
    }
    return (OM_STATUS_FAIL);
  }
  return (0);
}

int
icmp_checksum (struct icmp *icp)
{
  register u_short *w;
  register int sum;
  register int cc;
  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum), we add
   * sequential 16 bit words to it, and at the end, fold back all the
   * carry bits from the top 16 bits into the lower 16 bits.
   */
  /* skip the ICMP portion */
  cc = DEFDATALEN + 8;
  w = (u_short *) icp;
  sum = 0;
  /* sum the words */
  while (cc > 1) {
    sum += *w++;
    cc -= 2;
  }
  /* sum any odd byte */
  if (cc == 1) {
    sum += *(u_char *) w;
  }
  /* add in the carries */
  sum = (sum >> 16) + (sum & 0xffff);
  /* add in any carry of adding in carries */
  sum += (sum >> 16);
  return (~sum);		/* we want the NOT */
}

int
icmp_send (int sd, char *buf, int bufsize, struct sockaddr *to)
{
  int i;

  if ((i =
       sendto (sd, buf, (DEFDATALEN + 8), 0, to,
	       sizeof (struct sockaddr))) != (DEFDATALEN + 8)) {
/*
      omLog (OM_LOG_INFO, "wrote %s %d chars, ret=%d", hp->h_name,
	       (DEFDATALEN + 8), i);
*/
    if (i < 0) {
      sys_errno = errno;
      omLog (OM_LOG_CRIT, "perror: %s", strerror (errno));
    }
    return (OM_STATUS_FAIL);
  }
/*
    omLog (OM_LOG_DEBUG, "Echo Request %s : %d data bytes\n",
	     inet_ntoa ((struct in_addr )to->sin_addr.s_addr),
	     DEFDATALEN);
  hnamebuf,inet_ntoa (*(struct in_addr *) &to->sin_addr.s_addr), DEFDATALEN);
*/
  return (i);
}

double
icmp_triptime (struct timeval *tv, struct timeval *tp)
{

  if ((tv->tv_usec -= tp->tv_usec) < 0) {
    --tv->tv_sec;
    tv->tv_usec += 1000000;
  }
  tv->tv_sec -= tp->tv_sec;
  /* milliseconds */
  /* return (tv.tv_sec * 1000 + (tv.tv_usec / 1000)); */
  /* 100ths of a millisecond */
  return (tv->tv_sec * 100000 + (tv->tv_usec / 10));
}

int
omNetCost (char *target)
{
  static int ident;		/* process id to identify our packets */
  int s;			/* socket file descriptor */
  struct sockaddr whereto;	/* who to echorequest */

  struct timeval tt, tv;
  struct hostent *hp;
  struct sockaddr_in *to;
  struct sockaddr_in from;
  int i, packlen;
  u_char *packet;

  struct iphdr *iph;
  register struct icmp *icp;
  long triptime = 0;
  int hlen;
  int sys_errno;

  int nreceived = 0;		/* # of packets we got back */

  int cc;

  size_t fromlen;

  u_char outpack[MAXPACKET];

  if (!
      (packet =
       malloc ((u_int) (packlen = DEFDATALEN + MAXIPLEN + MAXICMPLEN)))) {
    omLog (OM_LOG_WARNING, "out of memory");
    return (OM_STATUS_FAIL);
  }

  if ((strlen (target) < 1) || (strlen (target) > MAXHOSTNAMELEN)) {
    omLog (OM_LOG_WARNING, "invalid host name string length");
    return (OM_STATUS_FAIL);
  }

  if (!(hp = gethostbyname (target))) {
    omLog (OM_LOG_WARNING, "unknown host %s", target);
    return (OM_STATUS_FAIL);
  }

  icmp_socket (&s);

  if (fcntl (s, F_SETFL, O_NONBLOCK) < 0) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "fcntl\nperror: %s", strerror (errno));
    return (OM_STATUS_FAIL);
  }

  memset (&whereto, 0, sizeof (struct sockaddr));
  to = (struct sockaddr_in *) &whereto;
  to->sin_family = hp->h_addrtype;
  if (hp->h_length > (int) sizeof (to->sin_addr)) {
    hp->h_length = sizeof (to->sin_addr);
  }
  memcpy (&to->sin_addr, hp->h_addr, hp->h_length);
  /* strncpy (hnamebuf, hp->h_name, sizeof (hnamebuf) - 1); */

  /* pad unused area with its own offset */
  for (i = 8 + sizeof (struct timeval); i < DEFDATALEN; ++i)
    outpack[i] = i;


  ident = getpid () & 0xFFFF;

  icp = (struct icmp *) outpack;
  icp->icmp_type = ICMP_ECHO;
  icp->icmp_code = 0;
  icp->icmp_cksum = 0;
  icp->icmp_seq = 0;
  icp->icmp_id = ident;


  omLog (OM_LOG_DEBUG, "Echo Request %s : %d data bytes\n",
	 inet_ntoa (*(struct in_addr *) &to->sin_addr.s_addr), DEFDATALEN);

  gettimeofday ((struct timeval *) icp->icmp_data, NULL);
  /* compute ICMP checksum */
  icp->icmp_cksum = icmp_checksum (icp);

  icmp_send (s, (char *) &outpack, (DEFDATALEN + 8), &whereto);

/*
    if (fcntl (s, F_SETFL, O_NONBLOCK) < 0) {
      sys_errno = errno;
      omLog (OM_LOG_CRIT, "fcntl\nperror: %s",strerror(errno));
      return (OM_STATUS_FAIL);
    }
*/
  gettimeofday (&tt, (struct timezone *) NULL);

  fromlen = sizeof (from);

  while (nreceived < 1) {
    gettimeofday (&tv, (struct timezone *) NULL);
    if (tv.tv_sec > tt.tv_sec + MAXWAIT)
      break;

    if ((cc =
	 recvfrom (s, (char *) packet, packlen, 0,
		   (struct sockaddr *) &from, (uint *) & fromlen)) < 0) {
      if (errno == EINTR || errno == EAGAIN)
	continue;
      sys_errno = errno;
      omLog (OM_LOG_CRIT, "recvfrom \nperror: %s", strerror (errno));
      break;
    }

    gettimeofday (&tv, (struct timezone *) NULL);

    /* Check the IP header */
    iph = (struct iphdr *) packet;
    hlen = (iph->ihl) << 2;
    if (cc < DEFDATALEN + ICMP_MINLEN) {
      omLog (OM_LOG_WARNING, " short packet %s: (%d bytes)\n",
	     inet_ntoa (*(struct in_addr *) &from.sin_addr.s_addr), cc);
    }

    /* Now the ICMP part */
    cc -= hlen;
    icp = (struct icmp *) (packet + hlen);
    if ((icp->icmp_id == ident) && (icp->icmp_type == ICMP_ECHOREPLY)) {

      ++nreceived;
      triptime = icmp_triptime (&tv, (struct timeval *) &icp->icmp_data);
      omLog (OM_LOG_DEBUG,
	     "Echo Reply icmp_seq=%u ttl=%d bytes=%d %ld.%ldms\n",
	     icp->icmp_seq, iph->ttl, cc, (triptime / 100), (triptime % 100));
      /*icp->icmp_seq,iph->ttl,cc, triptime / 10, triptime % 10); */
    }
  }

  if (nreceived == 0) {
    omLog (OM_LOG_NOTICE, " No Echo Received");
    return (OM_STATUS_FAIL);
  }
  return (triptime < MAXINT ? (int) triptime : MAXINT);
}
