/*
* Copyright (c) 2005 by G4SRA
*
* Permission to use, copy and distribute this software is hereby granted
* under the terms of version 2 or any later version of the GNU General Public
* License, as published by the Free Software Foundation.
*
* THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
* WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
* FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
*/
/*
* Authors: G4SRA (cradford@bcs.org.uk)
*/

#include <syslog.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#include "omsyslog.h"
#include "omstatus.h"

void
omLogWhatLevel (void)
{
#ifdef _DEBUG_
  omLog (LOG_DEBUG, "Loglevel DEBUG");
  omLog (LOG_INFO, "Loglevel INFO");
  omLog (LOG_NOTICE, "Loglevel NOTICE");
  omLog (LOG_WARNING, "Loglevel WARNING");
  omLog (LOG_ERR, "Loglevel ERR");
  omLog (LOG_CRIT, "Loglevel CRIT");
  omLog (LOG_ALERT, "Loglevel ALERT");
  /*omLog (LOG_EMERG, "Loglevel EMERG"); too much console/email garbage */
  omLog (LOG_ALERT, "Loglevel EMERG (implied)");
#endif
}

void
omLogIncLevel (void)
{
  omLogSetMask (LOG_UPTO ((setlogmask (0) << 1) + 1));
}

void
omLogDecLevel (void)
{
  omLogSetMask (LOG_UPTO (setlogmask (0) >> 1));
}

int
omLogSetMask (int level)
{
#ifdef _DEBUG_
  if (level != setlogmask (0)) {
    omLogWhatLevel ();
    if (level < setlogmask (0)) {
      omLog (LOG_DEBUG, "Loglevel changing from %X to %X", setlogmask (0),
	     level);
    }
  }
#endif
  level = setlogmask (level);
#ifdef _DEBUG_
  if (level != setlogmask (0)) {
    if (setlogmask (0) > level) {
      omLog (LOG_DEBUG, "Loglevel changed from %X to %X", level,
	     setlogmask (0));
    }
    omLogWhatLevel ();
  }
#endif
  return (level);
}

void
omLogToggleDebug ()
{
  static int level, toggle = (-1);

  if (toggle < 0) {
    toggle = 0;
    level = omLogSetMask (LOG_UPTO (LOG_DEBUG));
  }
  else {
    toggle = (-1);
    omLogSetMask (level);
  }
}

/* Send a single line of 80 characters max to syslogd */
void
omLog (int level, const char *fmt, ...)
{
  int retn;
  va_list vl;
  char buf[255];
  char *ptr1, *ptr2;

  int my_errno = errno;

  va_start (vl, fmt);
  retn = vsnprintf (buf, sizeof (buf) - 1, fmt, vl);
  va_end (vl);

#if 0
  /* strip newlines off the end of the message */
  while (buf[strlen (buf) - 1] == '\n') {
    buf[strlen (buf) - 1] == '\0';
  }
#endif

#if 0
  /* strip newlines out of the message */
  ptr1 = buf;
  ptr2 = buf;
  while (*ptr2 != '\0') {
    while (*ptr2 == '\n') {
      ++ptr2;
    }
    *ptr1++ = *ptr2++;
  }
  *ptr1 = *ptr2;
#endif

  /* replace all but trailing newlines with something else (space) */
  /* Note this modifies the passed string! */
  ptr1 = buf;
  ptr2 = buf;
  while (*ptr2 != '\0') {
    if (*ptr2 == '\n') {
      while (*ptr2 == '\n') {
	++ptr2;
      }
      if (*ptr2 != '\0') {
	if (*ptr2 != '"') {
	  *ptr1++ = '"';
	}
      }
    }
    *ptr1++ = *ptr2++;
  }
  /* copy the end of string marker */
  *ptr1 = *ptr2;

#ifdef _ROBUST_
  if ((level == LOG_CRIT) && (errno != sys_errno || errno != my_errno))
    syslog (LOG_ALERT, "errno=%d:%s\nmy_errno=%d:%s\nsys_errno=%d:%s", errno,
	    strerror (errno), my_errno, strerror (my_errno), sys_errno,
	    strerror (sys_errno));
#endif

  if (retn < sizeof (buf)) {
    syslog (level, "%s", buf);
  }
  else {
    syslog (LOG_ALERT, "Log message truncated\n%s", buf);
  }
}

    /* EMERG ALERT CRIT ERR WARNING NOTICE INFO DEBUG */

void
omLogInitialise (int flags, const char *program_name)
{
  if (flags) {
    openlog (program_name, LOG_CONS | LOG_PID | LOG_PERROR, LOG_DAEMON);
    setlogmask (LOG_UPTO (LOG_INFO));
  }
  else {
    openlog (program_name, LOG_CONS | LOG_PID, LOG_DAEMON);
    setlogmask (LOG_UPTO (LOG_NOTICE));
  }
}

void
omLogFinalise (void)
{
  closelog ();
}
