/*
 * Copyright (c) 2002 by Louis Zechtzer
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: Louis Zechtzer (lou@clarity.net)
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

#ifndef _OM_SYSLOG_H
#define _OM_SYSLOG_H

#include <stdio.h>
#include <syslog.h>

/*
 * Define logging severity levels.  These values are based on 
 * those in syslog.h.
 * 
 * The low 4 bits are for finer grained debugging messages.
 */

/* first three bits absolute value */
#define OM_LOG_EMERG       LOG_EMERG	/* system is unusable */
#define OM_LOG_ALERT       LOG_ALERT	/* action must be taken immediately */
#define OM_LOG_CRIT        LOG_CRIT	/* critical conditions */
#define OM_LOG_ERR         LOG_ERR	/* error conditions */
#define OM_LOG_WARNING     LOG_WARNING	/* warning conditions */
#define OM_LOG_NOTICE      LOG_NOTICE	/* normal but significant condition */
#define OM_LOG_INFO        LOG_INFO	/* informational */
#define OM_LOG_DEBUG       LOG_DEBUG	/* debug-level messages */

#define OM_LOG_PERROR	0	/* merge this with crit */

extern void omLogWhatLevel (void);
extern int omLogSetMask (int);
extern void omLogToggleDebug (void);
extern void omLog (int type, const char *, ...);
extern void omLogInitialise (int method, const char *progname);
extern void omLogFinalise (void);

#endif /* __OMD_SYSLOG_H */
