/*
* Copyright (c) 2005 by G4SRA
*
* Permission to use, copy and distribute this software is hereby granted
* under the terms of version 2 or any later version of the GNU General Public
* License, as published by the Free Software Foundation.
*
* THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
* WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
* FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
*
* Author: G4SRA (cradford@bcs.org.uk)
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "omwakeup.h"
#include "omstatus.h"
#include "omsyslog.h"
#include "node.h"

#define WOLDATALEN (6+(6*16))

int
omWakeUp (unsigned char *macaddr)
{

  unsigned char buf[WOLDATALEN + 1];
  unsigned char *ptr = buf;
  int i;

  int sd;
  struct sockaddr dest;
  struct sockaddr_in *to;
  int val = 1;
  int j;

  /* 6 x 0xff */
  for (i = 0; i < 6; i++)
    *ptr++ = 0xff;

  /* 16 x macaddr */
/* BUG ? This does not work, memmove stops at the first '\0' character  */
/*
    for (i = 0; i < 6; i++)
      *ptr++ = macaddr [i];

    memmove(ptr,&buf[6],90);
*/

  for (j = 0; j < 16; j++)
    for (i = 0; i < 6; i++)
      *ptr++ = macaddr[i];

  /* Set up broadcast address */
  memset (&dest, 0, sizeof (dest));
  to = (struct sockaddr_in *) &dest;
  to->sin_family = AF_INET;
  to->sin_addr.s_addr = (unsigned long) 0xffffffff;;
  /* to->sin_port = net_mcast_port_nbo; */
  /* FIXME: what port should we use ? */
  to->sin_port = htons (9999);

  if ((sd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "omUDPxmit socket:%s", strerror (sys_errno));
    return (OM_STATUS_FAIL);
  }

  /* Set socket options */
  if (setsockopt (sd, SOL_SOCKET, SO_BROADCAST, (char *) &val, sizeof (val)) <
      0) {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "omUDPxmit setsockopt:%s", strerror (sys_errno));
    close (sd);
    return (OM_STATUS_FAIL);
  }

  /* Send the packet out */
  while (sendto (sd, buf, WOLDATALEN, 0, &dest, sizeof (dest)) < 0) {
    sys_errno = errno;
    if (sys_errno == EINTR) {
      continue;
    }
    omLog (OM_LOG_CRIT, "omUDPxmit sendto:%s", strerror (sys_errno));
    close (sd);
    return (OM_STATUS_FAIL);
  }

  close (sd);
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "WakeUp packet sent to %s:%d",
	 inet_ntoa (to->sin_addr), ntohs (to->sin_port));
#endif
  return (OM_STATUS_SUCCESS);
}


int
omWakeupNode (omd_node * n)
{
#ifdef _DEBUG_
  struct in_addr saddr;
  saddr.s_addr = n->s_addr;
  omLog (OM_LOG_DEBUG, "Waking node %s (%8x)", inet_ntoa (saddr),
	 ntohl (n->s_addr));
#endif

  return (omWakeUp (n->macaddr));
}

#ifdef _TESTWAKEUP_
int
main (int argc, char *argv[])
{
  unsigned char macaddr[6] = {
    /* 0x00,0x00,0x00,0x00,0x00,0x00 */
    0x70, 0x08, 0x33, 0x95, 0x40, 0x00
      /* 00:40:95:33:08:70 */
  };

  omLogInitialise (1, "testwakeup");
  omLogSetMask (255);
  omWakeUp (macaddr);
  omLogFinalise ();
  return (0);
}
#endif
