/*
 * Copyright (c) 2005 by G4SRA
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

/* Request Load Average can be stale */
/* Request Free Memory can be stale */
/* Request Cluster Size can be stale */

/* Request Architecture cannot be stale */
/* Request Speed cannot be stale */
/* Request Version cannot be stale */

#include "bequest.h"
#include "request.h"
#include "tcp.h"
#include "omsyslog.h"
#include "omstatus.h"
#include "sys.h"
#include "openmosix.h"
#include "node.h"
#include "omlocal.h"
#include "omnetcost.h"
#include "ommacaddr.h"
#include "omd.h"

#include <signal.h>
#include <sys/time.h>

int reqIPNumCPUs (int sd);
int reqIPLoadAvg (int sd);
int reqIPFreeMem (int sd);
int reqIPClusterSize (int sd);
int reqIPLifetime (int sd);
int reqIPOMVer (int sd);
int reqIPArch (int sd);
int reqIPSpeed (int sd);
int reqIPCPUGrunt (int sd);
int reqIPNetCost (int sd);
int reqIPMACAddr (int sd);
int reqIPFlags (int sd);
int reqIPPowerDown (int sd);

int reqNodeNumIP (int sd);
int reqFubar (int sd);


static int nn;
static char ip[32] = { 0 };

/* order by most frequently used (volatile) first */
/* sizeof()-1 don't count the terminating '\0' */
reqfun_t reqfuntbl[] = {
  {MSG_REQ_IP_LOAD, (sizeof (MSG_REQ_IP_LOAD) - 1), reqIPLoadAvg, "%31s", ip}
  ,
  {MSG_REQ_IP_FREE_MEM, (sizeof (MSG_REQ_IP_FREE_MEM) - 1), reqIPFreeMem,
   "%31s", ip}
  ,
  {MSG_REQ_IP_ARCH, (sizeof (MSG_REQ_IP_ARCH) - 1), reqIPArch, "%31s", ip}
  ,
  {MSG_REQ_IP_CLUSTER_SIZE, (sizeof (MSG_REQ_IP_CLUSTER_SIZE) - 1),
   reqIPClusterSize, "%31s", ip}
  ,
  {MSG_REQ_IP_LIFE, (sizeof (MSG_REQ_IP_LIFE) - 1), reqIPLifetime, "%31s", ip}
  ,
  {MSG_REQ_IP_SPEED, (sizeof (MSG_REQ_IP_SPEED) - 1), reqIPSpeed, "%31s", ip}
  ,
  {MSG_REQ_IP_CPUGRUNT, (sizeof (MSG_REQ_IP_CPUGRUNT) - 1), reqIPCPUGrunt,
   "%31s", ip}
  ,
  {MSG_REQ_IP_MACADDR, (sizeof (MSG_REQ_IP_MACADDR) - 1), reqIPMACAddr,
   "%31s", ip}
  ,
  {MSG_REQ_IP_NETCOST, (sizeof (MSG_REQ_IP_NETCOST) - 1), reqIPNetCost,
   "%31s", ip}
  ,
  {MSG_REQ_IP_FLAGS, (sizeof (MSG_REQ_IP_FLAGS) - 1), reqIPFlags, "%31s", ip}
  ,
  {MSG_REQ_IP_NCPU, (sizeof (MSG_REQ_IP_NCPU) - 1), reqIPNumCPUs, "%31s", ip}
  ,
  {MSG_REQ_NN_NODE_IP, (sizeof (MSG_REQ_NN_NODE_IP) - 1), reqNodeNumIP, "%d",
   &nn}
  ,
  {MSG_REQ_IP_POWERDOWN, (sizeof (MSG_REQ_IP_POWERDOWN) - 1), reqIPPowerDown,
   "%31s", ip}
  ,
  {MSG_REQ_IP_OMV, (sizeof (MSG_REQ_IP_OMV) - 1), reqIPOMVer, "%31s", ip}
  ,
  {"FUBAR ", (sizeof ("FUBAR ") - 1), reqFubar}
};

int
omd_find_node_by_netaddr (const struct in_addr *node_addr)
{
  int i;
  uint32_t t_addr = node_addr->s_addr;

  for (i = 0; i < cluster->end; i++) {
    if ((uint32_t) v_get (cluster, i)->s_addr == t_addr) {
      /*
         omLog (OM_LOG_INFO, "Identified node %d %s %08x", i,
         inet_ntoa (*node_addr), ntohl (t_addr));
       */
      return (i);
    }
  }
  omLog (OM_LOG_NOTICE, "No such node %s (%08x)",
	 inet_ntoa (*node_addr), ntohl (t_addr));
  return (-1);
}

int
omd_find_node_by_ip (char *ip)
{
  struct in_addr addr;

  inet_pton (AF_INET, ip, &addr);
  return (omd_find_node_by_netaddr (&addr));
}

int
reqFubar (int sd)
{
  omLog (OM_LOG_ALERT, "Received Fubar request");
  return (OM_STATUS_FAIL);
}

/* FIXME: think about this a bit harder */
int
reqIPPowerDown (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    if (nn != 0) {
      char *rbuf;
      rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_POWERDOWN "0.0.0.0\n");
      if (rbuf != NULL) {
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      /* FIXME: the runlevel argument should be a "this node" parameter */
      tcp_send_selff (MSG_BEQ_NA_OMV "%08x 0\n", addr.s_addr);
      tcp_send (&sd, "POWERINGDOWN");
      return (OM_STATUS_SUCCESS);
    }
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "ERROR");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPOMVer (int sd)
{
  struct in_addr addr;
  int nn;
  char *ptr;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    char *rbuf;
    omd_node *n = v_get (cluster, nn);
    if ((ptr = get_node_omver (n)) != NULL)
      return (tcp_send (&sd, ptr));

    if (nn != 0)
      rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_OMV "0.0.0.0\n");
    else
      rbuf = getLocalOMLibVer ();

    if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
      tcp_send_selff (MSG_BEQ_NA_OMV "%08x %s\n", addr.s_addr, rbuf);
#else
      set_node_omver (rbuf);
#endif
      tcp_send (&sd, rbuf);
      free (rbuf);
      return (OM_STATUS_SUCCESS);
    }
    /* FIXME: we need to return somnething sensible */
    tcp_send (&sd, "ERROR");
    return (OM_STATUS_FAIL);
  }
  /* FIXME: we need to return somnething sensible */
  tcp_send (&sd, "ERROR");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPArch (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    char *rbuf;
    omd_node *n = v_get (cluster, nn);
    if (n->arch != NULL)
      return (tcp_send (&sd, n->arch));
/*
    if ((np=omNodeArchGet(nn))!=NULL)    
      return (tcp_send (&sd, np));
*/
    if (nn != 0)
      rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_ARCH "0.0.0.0\n");
    else
      rbuf = getLocalArch ();

    if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
      tcp_send_selff (MSG_BEQ_NA_ARCH "%08x %s\n", addr.s_addr, rbuf);
      tcp_send (&sd, rbuf);
      free (rbuf);
#else
      n->arch = rbuf;
      tcp_send (&sd, rbuf);
#endif
      return (OM_STATUS_SUCCESS);
    }
    /* FIXME: we need to return somnething sensible */
    tcp_send (&sd, "ERROR");
    return (OM_STATUS_FAIL);
  }
  /* FIXME: we need to return somnething sensible */
  tcp_send (&sd, "ERROR");
  return (OM_STATUS_BAD_NODEIP);
}

/*  depreciated - use effective CPUGrunt */
int
reqIPNumCPUs (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if (n->num_cpus >= 0)
      return (tcp_sendf (&sd, "%d", n->num_cpus));

    if (nn != 0) {
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_NCPU "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_NCPU "%08x %s\n", addr.s_addr, rbuf);
#else
	sscanf (rbuf, "%d", &n->num_cpus);
#endif
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      int numcpus = getLocalNumCPUs ();
      if (numcpus >= 0) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_NCPU "%08x %d\n", addr.s_addr, numcpus);
#else
	n->num_cpus = numcpus;
#endif
	tcp_sendf (&sd, "%d", numcpus);
	return (OM_STATUS_SUCCESS);
      }
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPCPUGrunt (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if (n->grunt >= 0)
      /* return the cached value */
      return (tcp_sendf (&sd, "%d", n->grunt));

    if (nn != 0) {
      /* Get cpu speed of remote node */
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_CPUGRUNT "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	/* Update our cache with the remote node data */
	tcp_send_selff (MSG_BEQ_NA_CPUGRUNT "%08x %s\n", addr.s_addr, rbuf);
#else
	sscanf (rbuf, "%d", &n->grunt);
#endif
	/* Forward the remote node data to the requesting node */
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      /* Get cpu speed of local node */
      int grunt = getLocalCPUGrunt ();
      if (grunt >= 0) {
#ifdef _FORKED_REQUESTER_
	/* Update our cache with the local node data */
	tcp_send_selff (MSG_BEQ_NA_CPUGRUNT "%08x %d\n", addr.s_addr, grunt);
#else
	n->grunt = grunt;
#endif
	/* Forward the local node data to the requesting node */
	tcp_sendf (&sd, "%d", grunt);
	return (OM_STATUS_SUCCESS);
      }
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

/* FIXME: We cannot calulate our own net cost, so we 
  cannot get the net cost of another node from that node.
  We must calculate the net cost for every node other than us, and
  average our own net cost from the results retrieved for us from
  all the other nodes.
 */
int
reqIPNetCost (int sd)
{
  struct in_addr addr;
  int nn;
  int netcost;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if (n->net_cost >= 0)
      return (tcp_sendf (&sd, "%d", n->net_cost));

    if (nn != 0) {
      netcost = omNetCost (ip);
      if (netcost >= 0) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_NETCOST "%08x %d\n", addr.s_addr, netcost);
#else
	n->net_cost = netcost;
#endif
	tcp_sendf (&sd, "%d", netcost);
	return (OM_STATUS_SUCCESS);
      }
    }
    /* We cannot calulate our own net cost */
    tcp_send (&sd, "-1");
    return (OM_STATUS_BAD_DATA);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPMACAddr (int sd)
{
  struct in_addr addr;
  char *macstr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if ((macstr = omMACAddrStr (n->macaddr)) != NULL)
      return (tcp_sendf (&sd, "%s", macstr));
/*
    if (nn != 0) {
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_MACADDR "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_MACADDR "%08x %s\n", addr.s_addr, rbuf);
#else
	 FIXME: 
#endif
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
*/
    tcp_send (&sd, "ERROR");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "ERROR");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPLifetime (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if (n->lifetime >= 0)
      return (tcp_sendf (&sd, "%d", n->lifetime));

    if (nn != 0) {
      /* Get parameter from remote node */
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_LIFE "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	/* Update our cache with the remote node data */
	tcp_send_selff (MSG_BEQ_NA_LIFE "%08x %s\n", addr.s_addr, rbuf);
#else
	sscanf (rbuf, "%d", &n->lifetime);
#endif
	/* Forward the remote node data to the requesting node */
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPFreeMem (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if (n->free_mem >= 0)
      return (tcp_sendf (&sd, "%d", n->free_mem));

    if (nn != 0) {
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_FREE_MEM "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_FREE_MEM "%08x %s\n", addr.s_addr, rbuf);
#else
	sscanf (rbuf, "%d", &n->free_mem);
#endif
	/* Forward the remote node data to the requesting node */
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      int freemem = getLocalFreeMem ();
      if (freemem >= 0) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_FREE_MEM "%08x %d\n", addr.s_addr,
			freemem);
#else
	n->free_mem = freemem;
#endif
	/* Forward the local node data to the requesting node */
	tcp_sendf (&sd, "%d", freemem);
	return (OM_STATUS_SUCCESS);
      }
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPLoadAvg (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if (n->load_avg >= 0)
      return (tcp_sendf (&sd, "%lf", n->load_avg));

    if (nn != 0) {
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_LOAD "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_LOAD "%08x %s\n", addr.s_addr, rbuf);
#else
	sscanf (rbuf, "%lf", &n->load_avg);
#endif
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      double loadavg = getLocalLoadAvg ();
      if (loadavg >= 0) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_LOAD "%08x %lf\n", addr.s_addr, loadavg);
#else
	n->load_avg = loadavg;
#endif
	tcp_sendf (&sd, "%lf", loadavg);
	return (OM_STATUS_SUCCESS);
      }
    }
    tcp_send (&sd, "-1.0");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "-1.0");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPFlags (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if (n->flags >= 0)
      return (tcp_sendf (&sd, "%d", n->flags));

    if (nn != 0) {
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_FLAGS "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_FLAGS "%08x %s\n", addr.s_addr, rbuf);
#else
	sscanf (rbuf, "%d", &n->flags);
#endif
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      if (sys_flags >= 0) {
#ifdef _FORKED_REQUESTER_
	tcp_send_selff (MSG_BEQ_NA_FLAGS "%08x %d\n", addr.s_addr, sys_flags);
#else
	n->flags = sys_flags;
#endif
	tcp_sendf (&sd, "%d", sys_flags);
	return (OM_STATUS_SUCCESS);
      }
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqIPClusterSize (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    if (nn == 0) {
      tcp_sendf (&sd, "%d", omd_cluster_size ());
      return (OM_STATUS_SUCCESS);
    }
/* FIXME: - CLUSTER SIZE ACCORDING TO REMOTE NODE */
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
/*
    omd_node *n = v_get (cluster, nn);
    if (n->num_nodes >=0)
      return (tcp_sendf (&sd, "%d", n->num_nodes));

    if (nn != 0) {
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_CLUSTER_SIZE "0.0.0.0\n");
      if (rbuf != NULL) {
	tcp_send_selff (MSG_BEQ_NA_CLUSTER_SIZE"%08x %s\n", addr.s_addr,
			rbuf);
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      int i = omd_cluster_size ();
      tcp_send_selff (MSG_BEQ_NA_CLUSTER_SIZE "%s%08x %d\n", addr.s_addr, i);
      tcp_sendf (&sd, "%d", i);
      return (OM_STATUS_SUCCESS);
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
  */
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

/* FIXME: - MIPS PER CPU */
/*  depreciated - use effective CPUGrunt */
int
reqIPSpeed (int sd)
{
  struct in_addr addr;
  int nn;

  inet_aton (ip, &addr);
  if ((nn = omd_find_node_by_netaddr (&addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
    if ((n->mips != NULL) && (n->mips[0] >= 0))
      return (tcp_sendf (&sd, "%d", n->mips[0]));

    if (nn != 0) {
      /* Get cpu speed of remote node */
      char *rbuf = tcp_exchange_msg (ip, MSG_REQ_IP_SPEED "0.0.0.0\n");
      if (rbuf != NULL) {
#ifdef _FORKED_REQUESTER_
	/* Update our cache with the remote node data */
	tcp_send_selff (MSG_BEQ_NA_SPEED "%08x %s\n", addr.s_addr, rbuf);
#else
	n->mips = omd_malloc (sizeof (int));
	sscanf (rbuf, "%d", &n->mips[0]);
#endif
	/* Forward the remote node data to the requesting node */
	tcp_send (&sd, rbuf);
	free (rbuf);
	return (OM_STATUS_SUCCESS);
      }
    }
    else {
      int *mips;
      /* Get cpu speed of local node */
      int i = getLocalCPUMips (&mips);
      if (i >= 0) {
#ifdef _FORKED_REQUESTER_
	/* Update our cache with the local node data */
	tcp_send_selff (MSG_BEQ_NA_SPEED "%08x %d\n", addr.s_addr, mips[0]);
#else
	n->mips = mips;
#endif
	/* Forward the local node data to the requesting node */
	tcp_sendf (&sd, "%d", mips[0]);
	return (OM_STATUS_SUCCESS);
      }
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_FAIL);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODEIP);
}

int
reqNodeNumIP (int sd)
{
  struct in_addr addr;

  if (nn < omd_cluster_size ()) {
    omd_node *n = v_get (cluster, nn);
    if (n->s_addr >= 0) {
      addr.s_addr = n->s_addr;
      return (tcp_send (&sd, inet_ntoa (addr)));
    }
    tcp_send (&sd, "-1");
    return (OM_STATUS_BAD_DATA);
  }
  tcp_send (&sd, "-1");
  return (OM_STATUS_BAD_NODENUM);
}

void
omd_non_volatile (const struct in_addr *addr)
{
  int sd1 = (-1), sd2 = (-1);
  char *rbuf = NULL;

#ifndef _FORKED_REQUESTER_
  if ((nn = omd_find_node_by_netaddr (addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
#endif

    if ((sd1 = tcp_chat_start (addr->s_addr)) >= 0) {
      if ((sd2 = tcp_chat_start (INADDR_ANY)) >= 0) {
	if (tcp_send (&sd1, MSG_REQ_IP_OMV "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_OMV "%08x %s\n", addr->s_addr, rbuf);
	    free (rbuf);
#else
	    set_node_omver (n, rbuf);
#endif
	  }
	}
	if (tcp_send (&sd1, MSG_REQ_IP_ARCH "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_ARCH "%08x %s\n", addr->s_addr, rbuf);
	    free (rbuf);
#else
	    set_node_arch (n, rbuf);
#endif
	  }
	}
	if (tcp_send (&sd1, MSG_REQ_IP_CPUGRUNT "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_CPUGRUNT "%08x %s\n", addr->s_addr,
		       rbuf);
#else
	    sscanf (rbuf, "%d", &n->grunt);
#endif
	    free (rbuf);
	  }
	}
	if (tcp_send (&sd1, MSG_REQ_IP_MACADDR "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_MACADDR "%08x %s\n", addr->s_addr,
		       rbuf);
#else
	    set_node_macaddr (n, rbuf);
#endif
	    free (rbuf);
	  }
	}
      }
    }
    if (sd2 >= 0)
      close (sd2);
    if (sd1 >= 0)
      close (sd1);
#ifndef _FORKED_REQUESTER_
  }
#endif
}

void
omd_volatile (const struct in_addr *addr)
{
  int sd1 = (-1), sd2 = (-1);
  char *rbuf = NULL;
  char *ip;
  int netcost;

#ifndef _FORKED_REQUESTER_
  if ((nn = omd_find_node_by_netaddr (addr)) >= 0) {
    omd_node *n = v_get (cluster, nn);
#endif

    if ((sd1 = tcp_chat_start (addr->s_addr)) >= 0) {
      if ((sd2 = tcp_chat_start (INADDR_ANY)) >= 0) {
	/* volatile data - MUST have life first */
	if (tcp_send (&sd1, MSG_REQ_IP_LIFE "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_LIFE "%08x %s\n", addr->s_addr, rbuf);
#else
	    sscanf (rbuf, "%d", &n->lifetime);
#endif
	    free (rbuf);
	  }
	}
	if (tcp_send (&sd1, MSG_REQ_IP_FLAGS "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_FLAGS "%08x %s\n", addr->s_addr,
		       rbuf);
#else
	    sscanf (rbuf, "%d", &n->flags);
#endif
	    free (rbuf);
	  }
	}
	if (tcp_send (&sd1, MSG_REQ_IP_FREE_MEM "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_FREE_MEM "%08x %s\n", addr->s_addr,
		       rbuf);
#else
	    sscanf (rbuf, "%d", &n->free_mem);
#endif
	    free (rbuf);
	  }
	}
	if (tcp_send (&sd1, MSG_REQ_IP_LOAD "0.0.0.0\n") >= 0) {
	  if ((rbuf = tcp_recv (&sd1)) != NULL) {
#ifdef _FORKED_REQUESTER_
	    tcp_sendf (&sd2, MSG_BEQ_NA_LOAD "%08x %s\n", addr->s_addr, rbuf);
#else
	    sscanf (rbuf, "%lf", &n->load_avg);
#endif
	    free (rbuf);
	  }
	}
	ip = omStrdup (inet_ntoa (*addr));
	netcost = omNetCost (ip);
#ifdef _FORKED_REQUESTER_
	tcp_sendf (&sd2, MSG_BEQ_NA_NETCOST "%08x %d\n", addr->s_addr,
		   netcost);
#else
	n->net_cost = netcost;
#endif
	free (ip);

/* No good - the MAC Address we see may be a router or gateway */
/*
	{
	  struct in_addr saddr;
	  unsigned char *macaddr;

	  saddr.s_addr = addr->s_addr;
	  macaddr = omGetMACAddr (inet_ntoa (saddr));
	}
*/
      }
    }
    if (sd2 >= 0)
      close (sd2);
    if (sd1 >= 0)
      close (sd1);
#ifndef _FORKED_REQUESTER_
  }
#endif
}

int
omd_complete_node (const struct in_addr *addr)
{
#ifdef _FORKED_REQUESTER_
  if (fork () == 0) {
#endif
    omd_non_volatile (addr);
    omd_volatile (addr);
    /* this is in reality our parent */
    kill (omd_this_pid, SIGUSR1);
#ifdef _FORKED_REQUESTER_
    exit (0);
  }
#endif
  return OM_STATUS_SUCCESS;
}

int
handle_request (int sd, char *buf)
{
  int i;

  for (i = 0; i < sizeof (reqfuntbl) / sizeof (reqfun_t); i++) {
    if (!strncmp (buf, reqfuntbl[i].msg, reqfuntbl[i].len)) {
      sscanf (&buf[reqfuntbl[i].len], reqfuntbl[i].fmt, reqfuntbl[i].arg1);
#ifdef _DEBUG_
      omLog (OM_LOG_DEBUG, "Received request \"%s\"", buf);
#endif
      return (reqfuntbl[i].fun (sd));
    }
  }
#ifdef _DEBUG_
  omLog (OM_LOG_DEBUG, "handle_request \"%s\" unknown", buf);
#endif
  return (OM_STATUS_BAD_REQ);
}
