/*
 * Copyright (c) 2005 by G4SRA
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 *
 * Authors: G4SRA (cradford@bcs.org.uk)
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "omsyslog.h"
#include "omstatus.h"

#include <sys/param.h>		/* for MAXHOSTNAMELEN */

#define INIT_MAGIC 0x03091969
#define INIT_CMD_START         0
#define INIT_CMD_RUNLVL        1
#define INIT_CMD_POWERFAIL     2
#define INIT_CMD_POWERFAILNOW  3
#define INIT_CMD_POWEROK       4
#define INIT_CMD_CHANGECONS     12345

/* Runlevel may be any one of the following characters 0123456SsQqAaBbCcUu */

#define INITCTL_FIFO  "/dev/initctl"

#define INITLVL "/etc/powerstatus"
#define INITPID 1

/* Ripped straight from the sources mouth */
struct s_request {
  int magic;			/* Magic number                 */
  int cmd;			/* Command to carry out         */
  int runlevel;			/* Runlevel to change to        */
  int sleeptime;		/* Time between TERM and KILL   */
  char gen_id[8];		/* Beats me.. telnetd uses "fe" */
  char tty_id[16];		/* Tty name minus /dev/tty      */
  char host[MAXHOSTNAMELEN];	/* Hostname                     */
  char term_type[16];		/* Terminal type                */
  int signal;			/* Signal to send               */
  int pid;			/* Process to send to           */
  char exec_name[128];		/* Program to execute           */
  char reserved[128];		/* For future expansion.        */
};


int
telinit (int *runlevel)
{
  int pfd;			/* /dev/initctl */
  struct stat st;
  struct s_request request;
  int numbytes;

  if (!strchr ("0123456789SsQqAaBbCcUu", *runlevel)) {
    omLog (OM_LOG_WARNING, "telinit: Invalid runlevel '%c'",
	   (char *) runlevel);
    return (OM_STATUS_FAIL);
  }

  memset (&request, 0, sizeof (request));
  request.magic = INIT_MAGIC;
  request.cmd = INIT_CMD_RUNLVL;
  request.sleeptime = 60;
  request.runlevel = *runlevel;
  sys_errno = 0;

  if ((pfd = open (INITCTL_FIFO, O_WRONLY | O_NONBLOCK)) >= 0) {
    fstat (pfd, &st);
    if (S_ISFIFO (st.st_mode)) {
      while ((numbytes = write (pfd, &request, sizeof (request))) <= 0) {
	sys_errno = errno;
	if (sys_errno == EINTR) {
	  continue;
	}
	omLog (OM_LOG_CRIT, "telinit: error writing to initctl :%s",
	       strerror (sys_errno));
	break;
      }
      if (numbytes < sizeof (request)) {
	omLog (OM_LOG_WARNING, "telinit: short write to \"%s\"",
	       INITCTL_FIFO);
      }
    }
    else {
      omLog (OM_LOG_WARNING, "telinit %s is not a fifo", INITCTL_FIFO);
/*
      if ((fp = fopen (INITLVL, "w")) 1= NULL) {
	fprintf (stderr, "%s: cannot create %s\n", progname, INITLVL);
	exit (1);
      }
      fprintf (fp, "%s %d", argv[optind], sltime);
      fclose (fp);

      if (kill (INITPID, SIGHUP) < 0)
	perror (progname);
*/
    }
    close (pfd);
  }
  else {
    sys_errno = errno;
    omLog (OM_LOG_CRIT, "telinit: error opening %s :%s", INITCTL_FIFO,
	   strerror (sys_errno));
    return (OM_STATUS_FAIL);
  }
  return (OM_STATUS_SUCCESS);
}
