/*
 * Copyright (c) 2005 by Nathan Wiegand 
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/* 
 * Authors: Nathan Wiegand (nathanwiegand@gmail.com)
 * Authors: G4SRA (cradford@bcs.org.uk)
 */

#include <math.h>

#include "vector.h"
#include "node.h"

inline unsigned int
log2 (unsigned int s)
{
  if (s <= 1)
    return 0;
  return log2 (s >> 1) + 1;
}

vector *
create (unsigned int size)
{
  vector *ret = NULL;

  size = 1 << log2 (size);

  if (size <= OM_MAX_CLUSTER_SIZE) {
    if ((ret = (vector *) malloc (sizeof (vector))) != NULL) {
      if ((ret->data =
	   (omd_node **) malloc (sizeof (omd_node *) * size)) != NULL) {
	ret->current_size = size;
	ret->end = 0;
	return ret;
      }
      free (ret);
    }
  }
  return NULL;
}

vector *
v_add (vector * vect, omd_node * node)
{

  if (vect->current_size < OM_MAX_CLUSTER_SIZE) {
    if (vect->end + 1 == vect->current_size) {
      int i;
      vector *newvector = create (vect->current_size * 2);

      if (newvector == NULL)
	return NULL;

      for (i = 0; i < vect->end; i++) {
	newvector->data[i] = vect->data[i];
      }
      newvector->end = vect->end;
      free (vect->data);
      free (vect);

      newvector->data[newvector->end++] = node;
      return newvector;
    }
    vect->data[vect->end++] = node;
    return vect;
  }
  return NULL;
}

/* FIXME: Hunt the bug, node destroyed, dangling vector still intact */
vector *
v_remuve (vector * vect, unsigned int position)
{

  if (vect->end < 1) {
    return NULL;
  }

  if (vect->end - 1 > vect->current_size / 4) {
    int i;

    for (i = position; i < vect->current_size - 1 && vect->data[i]; i++) {
      vect->data[i] = vect->data[i + 1];
    }
    vect->end--;
    return vect;
  }
  else {
    int i;
    vector *newvector = create (vect->current_size >> 1);

    if (newvector == NULL) {
      for (i = position; i < vect->current_size - 1 && vect->data[i]; i++) {
	vect->data[i] = vect->data[i + 1];
      }
      vect->end--;
      return vect;
    }

    for (i = 0; i < position; i++) {
      newvector->data[i] = vect->data[i];
    }
    /* skip position */
    ++i;
    for (; i < vect->end; i++) {
      newvector->data[i] = vect->data[i];
    }
    newvector->end = vect->end;
    newvector->end--;
    free (vect->data);
    free (vect);
    return newvector;
  }
  return NULL;
}

/* Rewrite :>) */
vector *
v_remove (vector * vect, unsigned int position)
{
  int i;
  vector *newvector;

  if (vect->end < 1) {
    return NULL;
  }

  if (vect->end - 1 > vect->current_size / 4) {
    for (i = position; i < vect->end - 1; i++) {
      vect->data[i] = vect->data[i + 1];
    }
    vect->end--;
    return vect;
  }
  else {
    newvector = create (vect->current_size >> 1);
    for (i = 0; i < position; i++) {
      newvector->data[i] = vect->data[i];
    }
    for (; i < vect->end - 1; i++) {
      newvector->data[i] = vect->data[i + 1];
    }
    newvector->end = vect->end - 1;
    free (vect->data);
    free (vect);
    return newvector;
  }
}

omd_node *
v_get (vector * vect, unsigned int position)
{
  if (position < vect->end) {
    return vect->data[position];
  }
  return NULL;
}

omd_node *
v_set (vector * vect, omd_node * node, unsigned int position)
{
  if (position < vect->end) {
    vect->data[position] = node;
    return vect->data[position];
  }
  return NULL;
}

int
v_swap (vector * vect, unsigned int pos1, unsigned int pos2)
{
  omd_node *ptmp;

  if ((pos1 < vect->end) && (pos2 < vect->end)) {
    ptmp = vect->data[pos1];
    vect->data[pos1] = vect->data[pos2];
    vect->data[pos2] = ptmp;
    return (0);
  }
  return (-1);
}
